#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <regex.h>
#include <string.h>
#include <errno.h>

#include "udm_common.h"
#include "udm_db.h"
#include "udm_charset.h"
#include "udm_proto.h"
#include "udm_parseurl.h"
#include "udm_parser.h"
#include "udm_spell.h"
#include "udm_conf.h"
#include "udm_indexer.h"
#include "udm_log.h"
#include "udm_hrefs.h"
#include "udm_robots.h"
#include "udm_utils.h"

static char _extra_headers[UDMSTRSIZ]="";
static char _user_agent[UDMSTRSIZ]="";

char * ExtraHeaders	=	_extra_headers;
char * UserAgent	=	_user_agent;
int    MaxDocSize	=	MAXDOCSIZE;


/* One server structure to be available on all config levels*/
/* to pass parameters */

static UDM_SERVER	*	csrv=NULL;
static UDM_SERVER	*	Server=NULL;
static UDM_FILTER	*	Filter=NULL;
static UDM_ALIAS	*	Alias=NULL;

static int nservers=0,mservers=0;
static int nfilters=0,mfilters=0;
static int naliases=0,maliases=0;

static char conf_err_str[UDMSTRSIZ]="";



/************ Allow/Disallow/CheckOnly  stuff *****************/


static int AddFilter(char * filter,int filter_type,int reverse){
#define ERRSTRSIZE 100
char regerrstr[UDMSTRSIZ]="";
int err;
	if(nfilters>=mfilters){
		if(mfilters){
			mfilters+=16;
			Filter=(UDM_FILTER *)realloc(Filter,mfilters*sizeof(UDM_FILTER));
		}else{
			mfilters=16;
			Filter=(UDM_FILTER *)malloc(mfilters*sizeof(UDM_FILTER));
		}
	}
	err=regcomp(&(Filter[nfilters].filter),filter,REG_EXTENDED|REG_ICASE);
	if(err){
		regerror(err, &(Filter[nfilters].filter), regerrstr, ERRSTRSIZE);
		sprintf(conf_err_str,"Wrong regex in config file: %s: %s", filter,regerrstr);
		regfree(&(Filter[nfilters].filter));
		return(1);
	}
	Filter[nfilters].filter_type=filter_type;
	Filter[nfilters].reverse=reverse;
	Filter[nfilters].regstr=strdup(filter);
	nfilters++;
	return(0);
}
static void FreeFilters(){
int i;
	for(i=0; i < nfilters; i++){
		regfree(&(Filter[i].filter));
		free(Filter[i].regstr);
	}
	free_it(Filter);
	nfilters=0;
	mfilters=0;
}

static char FCBuf[64]="";
char *FilterComStr(UDM_FILTER* f){
	FCBuf[0]=0;
	if(!f)return(FCBuf);
	switch(f->filter_type){
		case UDM_ALLOW: sprintf(FCBuf,"Allow");break;
		case UDM_DISALLOW: sprintf(FCBuf,"Disallow");break;
		case UDM_HEAD: sprintf(FCBuf,"CheckOnly");break;
		default: sprintf(FCBuf,"Unknown");break;
	}
	sprintf(STREND(FCBuf),"%s %s",
		(f->reverse?"NoMatch":""),
		(f->regstr?f->regstr:""));
	return(FCBuf);
}
UDM_FILTER * FindFilter(char *url){
int i;
#define NS 10
regmatch_t subs[NS];
int err;
	for(i=0;i<nfilters;i++){
		err=regexec(&(Filter[i].filter),url,NS,subs,0);
		if((Filter[i].reverse)&&(err))
			return(&Filter[i]);
		if((!Filter[i].reverse)&&(!err))
			return(&Filter[i]);
	}
	return(NULL);
}



/******************** ALIASES ***************************************/

static int AddAlias(char * find,char *replace){
	if(naliases>=maliases){
		if(maliases){
			maliases+=16;
			Alias=(UDM_ALIAS *)realloc(Alias,maliases*sizeof(UDM_ALIAS));
		}else{
			maliases=16;
			Alias=(UDM_ALIAS *)malloc(maliases*sizeof(UDM_ALIAS));
		}
	}
	Alias[naliases].find=strdup(find);
	Alias[naliases].replace=strdup(replace);
	naliases++;
	return(0);
}
static void FreeAliases(){
int i;
	for(i=0;i<naliases;i++){
		free(Alias[i].find);
		free(Alias[i].replace);
	}
	naliases=0;
	maliases=0;
	free_it(Alias);
}

UDM_ALIAS * FindAlias(char * str){
int j;
	for(j=0;j<naliases;j++){
		if(!STRNCMP(str,Alias[j].find)){
			return(&Alias[j]);
		}
	}
	return(NULL);
}


/*************************** Servers *************************************/

#define DEFAULT_PROXY_PORT	3128

static int AddServer(UDM_SERVER * srv){

	if(nservers>=mservers){
		if(mservers){
			mservers+=16;
			Server=(UDM_SERVER *)realloc(Server,mservers*sizeof(UDM_SERVER));
		}else{
			mservers=16;
			Server=(UDM_SERVER *)malloc(mservers*sizeof(UDM_SERVER));
		}
	}
	Server[nservers].url=strdup(srv->url);
	Server[nservers].charset=srv->charset?strdup(srv->charset):NULL;
	Server[nservers].basic_auth=srv->basic_auth?strdup(srv->basic_auth):NULL;
	Server[nservers].htdb_list=srv->htdb_list?strdup(srv->htdb_list):NULL;
	Server[nservers].htdb_doc=srv->htdb_doc?strdup(srv->htdb_doc):NULL;
	if(srv->proxy){
		char *s;
		Server[nservers].proxy=strdup(srv->proxy);
		if((s=strchr(Server[nservers].proxy,':'))){
			*s=0;
			Server[nservers].proxy_port=atoi(s+1)?atoi(s+1):DEFAULT_PROXY_PORT;
		}else{
			Server[nservers].proxy_port=DEFAULT_PROXY_PORT;
		}
	}else{
		Server[nservers].proxy=NULL;
		Server[nservers].proxy_port=0;
	}
	Server[nservers].period=srv->period;
	Server[nservers].hint=srv->hint;
	Server[nservers].net_errors=0;
	Server[nservers].outside=srv->outside;
	Server[nservers].maxhops=srv->maxhops;
	Server[nservers].gindex=srv->gindex;
	Server[nservers].gfollow=srv->gfollow;
	Server[nservers].deletebad=srv->deletebad;
	Server[nservers].userobots=srv->userobots;
	Server[nservers].bodyweight=srv->bodyweight;
	Server[nservers].titleweight=srv->titleweight;
	Server[nservers].urlweight=srv->urlweight;
	Server[nservers].descweight=srv->descweight;
	Server[nservers].keywordweight=srv->keywordweight;
	Server[nservers].max_net_errors=srv->max_net_errors;
	Server[nservers].read_timeout=srv->read_timeout;
	Server[nservers].delete_no_server=srv->delete_no_server;
	Server[nservers].correct_factor=srv->correct_factor;
	Server[nservers].incorrect_factor=srv->incorrect_factor;
	Server[nservers].number_factor=srv->number_factor;
	Server[nservers].alnum_factor=srv->alnum_factor;
	Server[nservers].min_word_length=srv->min_word_length;
	Server[nservers].max_word_length=srv->max_word_length;
	nservers++;
	return(0);
}
static void FreeServers(){
int i;
	for(i=0;i<nservers;i++){
		free_it(Server[i].url);
		free_it(Server[i].charset);
		free_it(Server[i].basic_auth);
		free_it(Server[i].htdb_list);
		free_it(Server[i].htdb_doc);
		free_it(Server[i].proxy);
	}
	nservers=mservers=0;
	free_it(Server);
}
UDM_SERVER * FindServer(char *url){
int i;
char *r;

	if(!Server)return(NULL);
	url=remove2dot(url);
	r=strstr(url,"/robots.txt");
	/* If it's a robot.txt, cut to hostinfo and find result */
	if((r)&&(!strcmp(r,"/robots.txt"))){
		r++;*r=0;
		for(i=0;i<nservers;i++){
			if(!STRNCMP(url,Server[i].url)){
				*r='r';
				return(&Server[i]);
			}
		}
		*r='r';
	}else{
		for(i=0;i<nservers;i++)
			if(!STRNCMP(url,Server[i].url))
				return(&Server[i]);
	}
	return(0);
}
static int cmpserver(const void *s1,const void *s2){
	return(strlen(((UDM_SERVER*)s2)->url)-strlen(((UDM_SERVER*)s1)->url));
}
static int InitServer(UDM_SERVER * srv){
	srv->url=NULL;
	srv->proxy=NULL;
	srv->basic_auth=NULL;
	srv->charset=NULL;
	srv->htdb_list=NULL;
	srv->htdb_doc=NULL;
	srv->proxy_port=0;
	srv->period=DEFAULT_REINDEX_TIME;
	srv->hint=0;
	srv->net_errors=0;
	srv->max_net_errors=MAXNETERRORS;
	srv->read_timeout=READ_TIMEOUT;
	srv->outside=0;
	srv->delete_no_server=1;
	srv->maxhops=DEFAULT_MAX_HOPS;
	srv->gindex=1;
	srv->gfollow=1;
	srv->deletebad=0;
	srv->userobots=1;
	srv->bodyweight=1;
	srv->titleweight=2;
	srv->keywordweight=2;
	srv->descweight=2;
	srv->urlweight=0;
	
	srv->correct_factor=1;
	srv->incorrect_factor=1;
	
	srv->number_factor=1;
	srv->alnum_factor=1;
	srv->min_word_length=1;
	srv->max_word_length=32;
	return(0);
}

__INDLIB__ char * ConfErrMsg(){
	return(conf_err_str);
}

/****************************  Load Configuration **********************/


#define IF_NUM_PARAM(src,name,dst,def) \
if(!STRNCASECMP(src,name)){if(sscanf(src+strlen(name),"%d",&(dst))!=1)dst=def;}

#define IF_BOOL_PARAM(src,name,dst,def) \
if(!STRNCASECMP(str,name)){ \
char xxx[1024]; \
if(sscanf(str+7,"%s",xxx)==1){ \
	if(!STRNCASECMP(xxx,"yes"))dst=1; \
	else	dst=0; \
}else \
	dst=def; \
}


__INDLIB__ int LoadConfig(char *conf_name,int config_level,int load_flags){
int i=0; /* For line number */
FILE *config;
char str[UDMSTRSIZ]="";
char str1[UDMSTRSIZ]="";
char str2[UDMSTRSIZ]="";
int has_dict=0;


	if(config_level==0){ /* Do some initialization */
		csrv=(UDM_SERVER*)malloc(sizeof(UDM_SERVER));
		InitServer(csrv);
		SetMaxURLNumber(-1);
		sprintf(UserAgent,"%s/%s",UDM_USER_AGENT,VERSION);
		ExtraHeaders[0]=0;
		MaxDocSize=MAXDOCSIZE;
		/*GMToffs=get_gmtoff();*/

		SetLogFacility("");

		Force1251=0;
		DBPort=0;
		free_it(DBHost);
		free_it(DBName);
		free_it(DBUser);
		free_it(DBPass);
		free_it(LocalCharset);
		LCindex=0;
		makeseparator(SEPARATOR,LCindex);
		FreeFilters();
		FreeAliases();
		free_it(MirrorRoot);
		free_it(MirrorHeadersRoot);
		use_mirror = UDM_MIRROR_NO;

		/* I have removed it to allow adding of URLs  */
		/* Passed in -i with -u command line args     */
		/* But it seems that UnloadConf() is required */
		/* REMOVED      FreeHrefs(); */

#ifdef USE_FILE
		FreeTypes();
#endif
		FreeServers();
	}

	/* Open config file */
	if(!(config=fopen(conf_name,"r"))){
		sprintf(conf_err_str,"Error: can't open config file '%s': %s",conf_name, strerror(errno));
		return(1);
	}

	/*  Read lines and parse */
	while(fgets(str1,sizeof(str1),config)){
		char *end;
		i++;
		if(!str1[0])continue;
		end=str1+strlen(str1)-1;
		while((end>=str1)&&(*end=='\r'||*end=='\n')){
			*end=0;if(end>str1)end--;
		}
		if(!str1[0])continue;
		if(str1[0]=='#')continue;

		if(*end=='\\'){
			*end=0;strcat(str,str1);
			continue;
		}
		strcat(str,str1);
		strcpy(str1,"");

		if(!STRNCASECMP(str,"DBHost")){
			if(sscanf(str+6,"%s",str1))
				DBHost=strdup(str1);
		}else
		if(!STRNCASECMP(str,"DBName")){
			DBName=strdup(trim_chars(str+6," \r\n\t"));
		}else
		if(!STRNCASECMP(str,"DBUser")){
			if(sscanf(str+6,"%s",str1)&&strcmp(str1,""))
				DBUser=strdup(str1);
		}else
		if(!STRNCASECMP(str,"DBPass")){
			if(sscanf(str+6,"%s",str1)&&strcmp(str1,""))
				DBPass=strdup(str1);
			else
				DBPass=strdup("");
		}else
		IF_NUM_PARAM(str,"DBPort",DBPort,0)
		else
		if(!STRNCASECMP(str,"LocalCharset")){
			if(sscanf(str+13,"%s",str1)&&strcmp(str1,"")){
				free_it(LocalCharset);
				LocalCharset=strdup(str1);
				LCindex=getcharset(LocalCharset);
				makeseparator(SEPARATOR,LCindex);
			}
		}else
		if(!STRNCASECMP(str,"DBType")){
			if(sscanf(str+6,"%s",str1)){
				if(!STRNCASECMP(str1,"msql"))DBType=UDM_DB_MSQL;
				else
				if(!STRNCASECMP(str1,"solid"))DBType=UDM_DB_SOLID;
				else
				if(!STRNCASECMP(str1,"oracle"))DBType=UDM_DB_ORACLE;
				else
					DBType = UDM_DB_UNK;
			}else{
				DBType = UDM_DB_UNK;
			}
		}else
		if(!STRNCASECMP(str,"DBMode")){
			if(sscanf(str+6,"%s",str1)){
				if(!STRNCASECMP(str1,"single"))DBMode=UDM_DBMODE_SINGLE;
				else
				if(!STRNCASECMP(str1,"crc"))DBMode=UDM_DBMODE_SINGLE_CRC;
				else
				if(!STRNCASECMP(str1,"multi"))DBMode=UDM_DBMODE_MULTI;
				else
					DBType = UDM_DBMODE_SINGLE;
			}else{
				DBType = UDM_DBMODE_SINGLE;
			}
		}else
		if(!STRNCASECMP(str,"DisallowNoMatch")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,UDM_DISALLOW,1))
					return(1);
			}
		}else
		if(!STRNCASECMP(str,"Disallow")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,UDM_DISALLOW,0))
					return(1);
			}
		}else
		if(!STRNCASECMP(str,"AllowNoMatch")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,UDM_ALLOW,1))
					return(1);
			}
		}else
		if(!STRNCASECMP(str,"Allow")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,UDM_ALLOW,0))
					return(1);
			}
		}else
		if(!STRNCASECMP(str,"CheckOnlyNoMatch")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,UDM_HEAD,1))
					return(1);
			}
		}else
		if(!STRNCASECMP(str,"CheckOnly")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,UDM_HEAD,0))
					return(1);
			}
		}else
#ifdef USE_FILE
		if(!STRNCASECMP(str,"AddType")){
			char *s,*s1,*lt;
			s=gettoken(str," \t\r\n",&lt);
			if((s1=gettoken(NULL," \t\r\n",&lt)))
				while((s=gettoken(NULL," \t\r\n",&lt)))
					AddType(s1,s,conf_err_str);
		}else
#endif
		if(!STRNCASECMP(str,"CharSet")){
			free_it(csrv->charset);
			if(sscanf(str+7,"%s",str1))
				csrv->charset=strdup(str1);
		}else
		if(!STRNCASECMP(str,"Proxy")){
			free_it(csrv->proxy);
			if(sscanf(str+5,"%s",str1)&&strcmp(str1,""))
				csrv->proxy=strdup(str1);
		}else
		if(!STRNCASECMP(str,"HTTPHeader")){
			if(sscanf(str+11,"%[^\n\r]s",str1)&&strcmp(str1,"")){
				if(!STRNCMP(str1,"User-Agent: ")){
					strcpy(UserAgent,str1+12);
				}else{
					strcat(ExtraHeaders,str1);
					strcat(ExtraHeaders,"\r\n");
				}
			}
		}else
#ifdef HAVE_SYSLOG_H
		if(!STRNCASECMP(str,"SyslogFacility")){
			if(sscanf(str+14,"%s",str1)&&strcmp(str1,""))
				SetLogFacility(str1);
		}else
#endif
		if(!STRNCASECMP(str,"AuthBasic")){
			free_it(csrv->basic_auth);
			if(sscanf(str+9,"%s",str1)){
				csrv->basic_auth=(char*)malloc(BASE64_LEN(strlen(str1)));
				base64_encode(str1,csrv->basic_auth,strlen(str1));
			}
		}else
		if(!STRNCASECMP(str,"HTDBList")){
			free_it(csrv->htdb_list);
			csrv->htdb_list=strdup(trim_chars(str+8," \t\r\n"));
		}else
		if(!STRNCASECMP(str,"HTDBDoc")){
			free_it(csrv->htdb_doc);
			csrv->htdb_doc=strdup(trim_chars(str+8," \t\r\n"));
		}else
		if(!STRNCASECMP(str,"Mime")){
			AddMime(str+4);
		}else
		if(!STRNCASECMP(str,"Alias")){
			if(sscanf(str+5,"%s%s",str1,str2)==2)
				AddAlias(str1,str2);
			else{
				sprintf(conf_err_str,"Error in config file '%s' line %d:%s",conf_name,i,str);
				return(1);
			}
		}else
		if(!STRNCASECMP(str,"Server")){
			UDM_URL from;
			if(sscanf(str+6,"%s",str1)!=1)
				continue;
			if(parseurl(&from,str1)){
				sprintf(conf_err_str,"File '%s' line %d: Invalid URL: %s",conf_name,i,str1);
				return(1);
			}
			csrv->url=strdup(str1);
			AddServer(csrv);
			if((!strcmp(from.schema,"http"))&&(load_flags&UDM_FLAG_ADD_SERV)&&(csrv->userobots)){
				sprintf(str,"%s://%s/%s",from.schema,from.hostinfo,"robots.txt");
				AddHref(NULL,str,0,0,0);
			}
			if(load_flags&UDM_FLAG_ADD_SERV)
				AddHref(NULL,str1,0,0,0);
		}
		else	IF_BOOL_PARAM(str,"FollowOutside",csrv->outside,0)
		else	IF_BOOL_PARAM(str,"Index",csrv->gindex,1)
		else	IF_BOOL_PARAM(str,"Follow",csrv->gfollow,1)
		else	IF_BOOL_PARAM(str,"Robots",csrv->userobots,1)
		else	IF_BOOL_PARAM(str,"DeleteBad",csrv->deletebad,0)
		else	IF_BOOL_PARAM(str,"DeleteNoServer",csrv->delete_no_server,1)
		else	IF_BOOL_PARAM(str,"ForceIISCharset1251",Force1251,0)

		else	IF_NUM_PARAM(str,"Tag",csrv->hint,0)
		else	IF_NUM_PARAM(str,"MaxNetErrors",csrv->max_net_errors,MAXNETERRORS)
		else	IF_NUM_PARAM(str,"ReadTimeOut",csrv->read_timeout,READ_TIMEOUT)
		else	IF_NUM_PARAM(str,"Period",csrv->period,DEFAULT_REINDEX_TIME)
		else	IF_NUM_PARAM(str,"MaxHops",csrv->maxhops,0)
		else	IF_NUM_PARAM(str,"BodyWeight",csrv->bodyweight,1)
		else	IF_NUM_PARAM(str,"TitleWeight",csrv->titleweight,2)
		else	IF_NUM_PARAM(str,"DescWeight",csrv->descweight,2)
		else	IF_NUM_PARAM(str,"UrlWeight",csrv->urlweight,0)
		else	IF_NUM_PARAM(str,"IspellCorrectFactor",csrv->correct_factor,1)
		else	IF_NUM_PARAM(str,"IspellIncorrectFactor",csrv->incorrect_factor,1)
		else	IF_NUM_PARAM(str,"NumberFactor",csrv->number_factor,1)
		else	IF_NUM_PARAM(str,"AlnumFactor",csrv->alnum_factor,1)
		else	IF_NUM_PARAM(str,"MinWordLength",csrv->min_word_length,1)
		else	IF_NUM_PARAM(str,"MaxWordLength",csrv->max_word_length,32)
		else	IF_NUM_PARAM(str,"KeywordWeight",csrv->keywordweight,2)
		else	IF_NUM_PARAM(str,"MaxDocSize",MaxDocSize,MAXDOCSIZE)
		else	if(!STRNCASECMP(str,"NoIndex"))csrv->gindex=0;
		else	if(!STRNCASECMP(str,"NoFollow"))csrv->gfollow=0;
		else
		if(!STRNCASECMP(str,"Affix")){
			char lang[20];
			if(load_flags&&UDM_FLAG_SPELL)
			if(2==sscanf(str+5,"%s%s",lang,str1)){
				if(*str1=='/')strcpy(str,str1);
				else	sprintf(str,"%s/%s",CONFIGDIR,str1);
				if(ImportAffixes(lang,str)){
					sprintf(conf_err_str,"Can't load affix :%s",str);
					return(1);
				}
			}else{
				sprintf(conf_err_str,"Error in config file '%s' line %d:%s",conf_name,i,str);
				return(1);
			}
		}else
		if(!STRNCASECMP(str,"Spell")){
			char lang[20];
			if(load_flags&&UDM_FLAG_SPELL)
			if(2==sscanf(str+5,"%s%s",lang,str1)){
				if(*str1=='/')strcpy(str,str1);
				else	sprintf(str,"%s/%s",CONFIGDIR,str1);
				/* -----  0 or 1 ?  */
				if(ImportDictionary(lang,str,0,"")){
					sprintf(conf_err_str,"Can't load dictionary :%s",str);
					return(1);
				}
				has_dict=1;
			}else{
				sprintf(conf_err_str,"Error in config file '%s' line %d:%s",conf_name,i,str);
				return(1);
			}
		}else
		if(!STRNCASECMP(str,"Include")){
			if((sscanf(str+7,"%s",str1))&&(config_level<5)){
				if(*str1=='/')strcpy(str,str1);
				else	sprintf(str,"%s/%s",CONFIGDIR,str1);
				LoadConfig(str,config_level+1,load_flags);
			}
		}else
		if(!STRNCASECMP(str,"MirrorRoot")){
			if((sscanf(str+10,"%s",str1))){
				if(*str1=='/')
					strcpy(str,str1);
				else
					sprintf(str,"%s/%s",CONFIGDIR,str1);
			} else {
				sprintf(str,"%s/mirrors",CONFIGDIR);
			}
			MirrorRoot = strdup(str);
			use_mirror=(use_mirror>0)?use_mirror:0;
		}else
		if(!STRNCASECMP(str,"MirrorHeadersRoot")){
			if((sscanf(str+17,"%s",str1))){
				if(*str1=='/')
					strcpy(str,str1);
				else
					sprintf(str,"%s/%s",CONFIGDIR,str1);
			} else {
				sprintf(str,"%s/headers",CONFIGDIR);
			}
			MirrorHeadersRoot = strdup(str);
			use_mirror=(use_mirror>0)?use_mirror:0;
		}else
		IF_NUM_PARAM(str,"MirrorPeriod",use_mirror,UDM_MIRROR_NO)
		else{
			sprintf(conf_err_str,"Error in config file line %d: %s",i,str);
			return(1);
		}
		str[0]=0;
	}
	fclose(config);


	/* On level0 : Free some variables, prepare others, etc */
	if(config_level==0){
		/* Add one virtual server if we want FollowOutside */
		/* or DeleteNoServer no	*/
		
		if((csrv->outside)||(!csrv->delete_no_server)){
			free_it(csrv->url);
			csrv->url="";
			AddServer(csrv);
		}else{
			free_it(csrv->url);
		}
		free_it(csrv->charset);
		free_it(csrv->proxy);
		free_it(csrv->basic_auth);
		free_it(csrv->htdb_list);
		free_it(csrv->htdb_doc);
		/*  Long name should be found first    */
		/*  to allow different options         */
		/*  for server and it's subdirectories */
		qsort((void*)Server,nservers,sizeof(UDM_SERVER),cmpserver);
		/* Sort ispell dictionay if it has been loaded */
		if(has_dict)SortDictionary();
	}
	return(0);
}
